/* ***** BEGIN LICENSE BLOCK *****
 *   Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is ThumbsDown.
 *
 * The Initial Developer of the Original Code is
 * Atte Kemppilä.
 * Portions created by the Initial Developer are Copyright (C) 2007
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 * 
 * ***** END LICENSE BLOCK ***** */


function thumbsdown_SqlExecutor(db, config, schema)
{
    this.config = this._normalizeConfig(config);
    this.db = this._openDatabase(db, schema);
    for (let [name, entry] in Iterator(config))
    {
        try
        {
            if ("stmt" in entry)
                config[name].stmt = this.db.createStatement(entry.stmt);
            else
                throw new Error("config entry has no statement (" + name + ")");
        }
        catch (e if (e.result === Components.results.NS_ERROR_FAILURE))
        {
            throw new Error("invalid sqlite statement: '" + entry.stmt + "' (" + name + ")");
        }
    }
}

thumbsdown_SqlExecutor.prototype =
{
    alias2sqlite:
    {
        int:        "Int32",
        int32:      "Int32",
        int64:      "Int64",
        double:     "Double",
        string:     "String",
        utf8string: "UTF8String",
        blob:       "Blob"
    },
    
    js2sqlite:
    {
        number: "Int32",
        string: "String"
    },
    
    execute: function(name, args)
    {
        args = this._normalizeArgs(Array.prototype.slice.call(arguments, 1));
        var entry = this._getConfigEntry(name);
        var stmt = entry.stmt;
        try
        {
            this._bindParameters(name, stmt, entry.params, args);
            stmt.execute();
            return this.db.lastInsertRowID;
        }
        finally
        {
            stmt.reset();
        }
    },
    
    executeStep: function(name, args)
    {
        args = this._normalizeArgs(Array.prototype.slice.call(arguments, 1));
        var entry = this._getConfigEntry(name);
        var stmt = entry.stmt;
        try
        {
            this._bindParameters(name, stmt, entry.params, args);
            while (stmt.executeStep())
            {
                var result;
                if (entry.single)
                    result = this._getStatementValue(stmt, 0, entry.result[0].type);
                else
                {
                    result = {};
                    if (entry.result.constructor === Array)
                        entry.result = this._normalizeResultConfig(entry.result, stmt);
                    var numEntries = stmt.numEntries;
                    for (var i = 0; i < numEntries; i++)
                    {
                        var colName = stmt.getColumnName(i);
                        var colType = entry.result[colName];
                        if (colType === undefined)
                            throw new Error("result type undefined for column '" + colName + "' (" + name + ")"); 
                        result[colName] = this._getStatementValue(stmt, i, colType);
                    }
                }
                yield result;
            }
        }
        finally
        {
            stmt.reset();
        }
    },
    
    fetchOne: function(name, args)
    {
        var result;
        var resultSet = this.executeStep.apply(this, arguments);
        try
        {
            result = resultSet.next();
            // only one result is expected so this should throw StopIteration
            resultSet.next();
            throw new Error("resultSet has more than one row: " + name + " " + args.toSource());
        }
        catch (e if (e instanceof StopIteration))
        {
            if (result === undefined)
                throw new Error("resultSet is empty: " + name + " " + args.toSource());
        }
        finally
        {
            resultSet.close();
        }
        return result;
    },
    
    fetchAll: function(name, args)
    {
        var resultSet = this.executeStep.apply(this, arguments);
        try
        {
            return [row for each (row in resultSet)];
        }
        finally
        {
            resultSet.close();
        }
    },
    
    _normalizeConfig: function(config)
    {
        for (let [name, entry] in Iterator(config))
        {
            // convert type aliases to proper sqlite types
            for (let [param, value] in Iterator(entry))
                switch (param)
                {
                    case "params":
                    case "result":
                        if (typeof(value) == "string")
                            entry[param] = value = value.toLowerCase().split(/[\s,]+/);
                        // loop that handles both arrays and objects
                        for each (index in Iterator(value, true))
                            if (value[index] in this.alias2sqlite)
                                value[index] = this.alias2sqlite[value[index]];
                            else
                                throw new Error("invalid sqlite type '" + value[index] + "' in '" + param +"' (" + name + ")");
                        break;
                }
            // default values for config parameters
            entry.params = (("params" in entry) ? entry.params : null);
            entry.single = (("single" in entry) ? entry.single : (entry.result && (entry.result.length == 1)));
        }
        return config;
    },
    
    _openDatabase: function(db, schema)
    {
        if (typeof(db) == "string")
        {
            var file = Components.classes["@mozilla.org/file/directory_service;1"].getService(Components.interfaces.nsIProperties).get("ProfD", Components.interfaces.nsIFile);
            file.append(db);
            db = file;
        }
        if (db instanceof Components.interfaces.nsIFile)
        {
            var exists = db.exists();
            var storageSvc = Components.classes["@mozilla.org/storage/service;1"].getService(Components.interfaces.mozIStorageService);
            db = storageSvc.openDatabase(db);
            if ((! exists) && schema)
                this._initDatabase(db, schema);
        }
        return db;
    },
    
    _initDatabase: function(db, schema)
    {
        db.beginTransaction();
        try
        {
            for each (stmt in schema)
                db.executeSimpleSQL(stmt);
            db.commitTransaction();
        }
        catch (e)
        {
            db.rollbackTransaction();
            throw e;
        }
    },
    
    _normalizeArgs: function(args)
    {
        if (args.length == 1)
        {
            var arg = args[0];
            if (typeof(arg) == "object")
                if (arg.constructor === Array)
                    return this._normalizeArgs(arg);
                else
                    return arg;
        }
        return args;
    },
    
    _getConfigEntry: function(name)
    {
        if (name in this.config)
            return this.config[name];
        else
            throw new Error("entry not found: " + name);
    },
    
    _bindParameters: function(name, stmt, params, args)
    {
        var isNamed = (args.constructor !== Array);
        for (let [key, value] in Iterator(args))
        {
            var index = (isNamed ? this._getStatementParameterIndex(stmt, key) : key);
            if (value != null)
            {
                var type = ((params != null) ? params[key] : this.js2sqlite[typeof(value)]);
                if (type === undefined)
                    throw new Error("invalid argument type: " + typeof(value) + " (" + value +  ") at '" + key + "' (" + name + ")");
                stmt["bind" + type + "Parameter"](index, value);
            }
            else
                stmt.bindNullParameter(index);
        }
    },
    
    _getStatementParameterIndex: function(stmt, name)
    {
        if (stmt.getParameterIndex)
            return stmt.getParameterIndex(":" + name);
        else
            // assuming that names are unique
            // https://bugzilla.mozilla.org/show_bug.cgi?id=388048 - see unit test for full usage example
            return stmt.getParameterIndexes(":" + name, {})[0] - 1;
    },
    
    _normalizeResultConfig: function(result, stmt)
    {
        var r = {};
        var numEntries = stmt.numEntries;
        for (var i = 0; i < numEntries; i++)
            r[stmt.getColumnName(i)] = result[i];
        return r;
    },
    
    _getStatementValue: function(stmt, index, type)
    {
        var value = null;
        if (! stmt.getIsNull(index))
            value = stmt["get" + type](index);
        return value;
    }
}
